# Importing modules
import random                                                                                                                                                                                                              
import pandas as pd        
import numpy as np 
import yfinance as yf
import datetime as dt                                                                                                      
from datetime import datetime  

RF = 0 # risk free rate
portfolio_returns = []
portfolio_risks = []
sharpe_ratios = []
portfolio_weights = []

# start_date = '2013-01-01' 
# end_date = '2025-03-27' 

# Object for creating an optimal portfolio                                                                                  
class EfPortfolio():                                                                                                        
    def __init__(self, symbols):                                                                                          
        self.assets = symbols                                   
    # Instantiating Variables                                                                                               
    def Vars(self, symbols, start_date, end_date):                                                                        
        self.tickers = symbols                                                                                  
        # Datetime                                                                                                          
        now = datetime.now()                                                                                                
        # Start date                                                                                                        
        self.start = start_date                                                                                           
        # End date                                                                                                          
        self.end= end_date          
        # Number of portfolios to create                                                                                 
        self.number_of_portfolios = 500                                                                                     
        # Create a Dataframe                                                                                                         
        self.data = pd.DataFrame()                                                                                          
        self.returns = pd.DataFrame() 
           
    # Storing results in a DataFrame                                                                                             
    def StockData(self):                                                                                                    
        for self.ticker in self.tickers:                                                                                    
            self.data[self.ticker] = yf.download(self.ticker, start=self.start, end=self.end, auto_adjust=False)['Adj Close'].pct_change()     
                                                                                                                                    
            if self.returns.empty:                                                                                          
                self.returns = self.data[[self.ticker]]                                                                     
            else:                                                                                                           
                self.returns = self.returns.join(self.data[[self.ticker]], how='outer')            
    # Calculating risk returns                                                                       
    def risk_reward(self):                                                                                                 
        for portfolio in range(self.number_of_portfolios):                                                                  
            #Generate random portfolio weights                                                                              
            self.weights = np.random.random_sample(len(self.tickers))                                                       
            self.weights = np.round((self.weights / np.sum(self.weights)), 3)                                               
            portfolio_weights.append(self.weights)                                                                
                                                                                                                                    
            #Calculate annualized returns                                                                                   
            self.annualized_return = np.sum(self.returns.mean() * self.weights) * 252                                       
            portfolio_returns.append(self.annualized_return)                                                                
                                                                                                                                    
            #Matrix covariance & portfolio risk calculation                                                                 
            self.matrix_covariance = self.returns.cov() * 252                                                               
            self.portfolio_variance = np.dot(self.weights.T,                                                                
                                            np.dot(self.matrix_covariance, self.weights))                                       
            self.portfolio_standard_deviation = np.sqrt(self.portfolio_variance)                                            
            portfolio_risks.append(self.portfolio_standard_deviation)                                                       
                                                                                                                                    
            #Sharpe ratio                                                                                                   
            self.sharpe_ratio = (self.annualized_return - RF) / self.portfolio_standard_deviation                           
            sharpe_ratios.append(self.sharpe_ratio)
    # Outputting optimization results                
    def results(self):
        self.portfolio_returns = np.array(portfolio_returns)
        self.portfolio_risks = np.array(portfolio_risks)
        self.sharpe_ratios = np.array(sharpe_ratios)

        self.portfolio_metrics = [self.portfolio_returns, self.portfolio_risks, self.sharpe_ratios, portfolio_weights]

        # Create DataFrame for Portfolio Metrics
        self.portfolios_df = pd.DataFrame(self.portfolio_metrics).T
        self.portfolios_df.columns = ['Returns', 'Risk', 'Sharpe Ratio', 'Weights']

        # Calculate Minimum Risk Portfolio
        self.min_risk = self.portfolios_df.iloc[self.portfolios_df['Risk'].astype(float).idxmin()]

        # Calculate Portfolio with Highest Return
        self.highest_return = self.portfolios_df.iloc[self.portfolios_df['Returns'].astype(float).idxmax()]

        # Calculate Portfolio with Highest Sharpe Ratio
        self.highest_sharpe = self.portfolios_df.iloc[self.portfolios_df['Sharpe Ratio'].astype(float).idxmax()] 

        # Portfolio yielding the lowest risk                                                                                                                 
        print('Lowest risk Allocation')                                                                                                
        print(self.min_risk)                                                                                                
        print(self.tickers)                                                                                                 
        print('')                                                                                                           
        # Portfolio yielding the highest                                                                                                                         
        print('Highest Returns Allocation')                                                                                            
        print(self.highest_return)                                                                                          
        print(self.tickers)                                                                                                 
        print('')                                                                                                           
        # Portfolio yielding the highest risk                                                                                                                         
        print('Highest Sharpe Allocation')                                                                                             
        print(self.highest_sharpe)                                                                                          
        print(self.tickers)                                                                                                 
        print('')


# Function for Calling Object
def risk_return(symbols, start_date, end_date):                                                                                                             
    # Instantiating Object                                                                                                        
    cEfPortfolio = EfPortfolio(symbols)                                                                                                                                                                         
    cEfPortfolio.Vars(symbols, start_date, end_date)                                                                                          
    cEfPortfolio.StockData()                                                                                                
    cEfPortfolio.risk_reward()                                                                                             
    cEfPortfolio.results()        
                                                                                
# risk_return(symbols)



# SPDX-FileCopyrightText: 2025-present zerubabeldinsa <zerubabel.dinsa@gmail.com>
#
# SPDX-License-Identifier: MIT

from .portfolio_construction.return_allocations.efficient_frontier import risk_return



import unittest
from unittest.mock import patch, MagicMock
import pandas as pd
import numpy as np
from trade_strategy.portfolio_construction.return_allocations.efficient_frontier import EfPortfolio, risk_return, portfolio_returns, portfolio_risks, sharpe_ratios, portfolio_weights

class TestEfficientFrontier(unittest.TestCase):

    def setUp(self):
        # Clear global lists before each test
        portfolio_returns.clear()
        portfolio_risks.clear()
        sharpe_ratios.clear()
        portfolio_weights.clear()

    @patch('yfinance.download')
    def test_risk_return_function(self, mock_yf_download):
        # Mock yfinance.download to return predictable data
        mock_data = pd.DataFrame({
            'Adj Close': [100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110]
        }, index=pd.to_datetime(pd.date_range(start='2023-01-01', periods=11, freq='D')))
        mock_yf_download.return_value = mock_data

        symbols = ['AAPL', 'GOOG']
        start_date = '2023-01-01'
        end_date = '2023-01-10'

        # Call the main function
        risk_return(symbols, start_date, end_date)

        # Assertions
        self.assertGreater(len(portfolio_returns), 0)
        self.assertGreater(len(portfolio_risks), 0)
        self.assertGreater(len(sharpe_ratios), 0)
        self.assertGreater(len(portfolio_weights), 0)

        self.assertEqual(len(portfolio_returns), len(portfolio_risks))
        self.assertEqual(len(portfolio_returns), len(sharpe_ratios))
        self.assertEqual(len(portfolio_returns), len(portfolio_weights))

        # Check if the results DataFrame is created and has expected columns
        # We need to instantiate EfPortfolio again to access its internal state after risk_return
        ef_portfolio_instance = EfPortfolio(symbols)
        ef_portfolio_instance.Vars(symbols, start_date, end_date)
        ef_portfolio_instance.StockData()
        ef_portfolio_instance.risk_reward()
        ef_portfolio_instance.results()

        self.assertIsInstance(ef_portfolio_instance.portfolios_df, pd.DataFrame)
        self.assertListEqual(list(ef_portfolio_instance.portfolios_df.columns), ['Returns', 'Risk', 'Sharpe Ratio', 'Weights'])
        self.assertFalse(ef_portfolio_instance.portfolios_df.empty)

        # Check if min_risk, highest_return, highest_sharpe are Series
        self.assertIsInstance(ef_portfolio_instance.min_risk, pd.Series)
        self.assertIsInstance(ef_portfolio_instance.highest_return, pd.Series)
        self.assertIsInstance(ef_portfolio_instance.highest_sharpe, pd.Series)

        

    @patch('yfinance.download')
    def test_ef_portfolio_stock_data(self, mock_yf_download):
        mock_data = pd.DataFrame({
            'Adj Close': [100, 101, 102, 103, 104, 105]
        }, index=pd.to_datetime(pd.date_range(start='2023-01-01', periods=6, freq='D')))
        mock_yf_download.return_value = mock_data

        symbols = ['MSFT']
        start_date = '2023-01-01'
        end_date = '2023-01-05'

        ef = EfPortfolio(symbols)
        ef.Vars(symbols, start_date, end_date)
        ef.StockData()

        self.assertIsInstance(ef.data, pd.DataFrame)
        self.assertIsInstance(ef.returns, pd.DataFrame)
        self.assertIn('MSFT', ef.data.columns)
        self.assertIn('MSFT', ef.returns.columns)
        self.assertFalse(ef.returns.empty)
        self.assertTrue(np.isnan(ef.returns['MSFT'].iloc[0])) # First value is NaN due to pct_change()
        self.assertAlmostEqual(ef.returns['MSFT'].iloc[1], 0.01) # (101-100)/100

    @patch('yfinance.download')
    def test_ef_portfolio_risk_reward_calculations(self, mock_yf_download):
        mock_data = pd.DataFrame({
            'Adj Close': [100, 101, 102, 103, 104, 105]
        }, index=pd.to_datetime(pd.date_range(start='2023-01-01', periods=6, freq='D')))
        mock_yf_download.return_value = mock_data

        symbols = ['MSFT']
        start_date = '2023-01-01'
        end_date = '2023-01-05'

        ef = EfPortfolio(symbols)
        ef.Vars(symbols, start_date, end_date)
        ef.StockData()
        ef.number_of_portfolios = 5 # Reduce for faster testing

        # Patch numpy.random.random_sample to return a fixed weight for predictability
        with patch('numpy.random.random_sample', return_value=np.array([1.0])):
            ef.risk_reward()

        self.assertEqual(len(portfolio_returns), ef.number_of_portfolios)
        self.assertEqual(len(portfolio_risks), ef.number_of_portfolios)
        self.assertEqual(len(sharpe_ratios), ef.number_of_portfolios)
        self.assertEqual(len(portfolio_weights), ef.number_of_portfolios)

        # Check if weights sum to 1 (approximately due to rounding)
        for weights in portfolio_weights:
            self.assertAlmostEqual(np.sum(weights), 1.0, places=3)

        # Basic check for non-zero values (since we mocked data, they should be calculable)
        self.assertTrue(all(r != 0 for r in portfolio_returns))
        self.assertTrue(all(r != 0 for r in portfolio_risks))
        # Sharpe ratio can be zero or negative, so not checking for non-zero

if __name__ == '__main__':
    unittest.main()



from efficient_frontier import risk_return

# Dates
start_date = '2013-01-01' 
end_date = '2025-03-27' 

# tickers
symbols = ["AAPL", "MSFT", "GOOG"]

# Calling function
risk_return(symbols, start_date, end_date)



# SPDX-FileCopyrightText: 2025-present zerubabeldinsa <zerubabel.dinsa@gmail.com>
#
# SPDX-License-Identifier: MIT
__version__ = "0.1.0"
__author__ = "zerubabeldinsa"
__email__ = "zerubabel.dinsa@gmail.com"
__license__ = "MIT"
__summary__ = "A Python package for analyzing investment portfolios, including using quantitative methods like the efficient frontier for calculations an optimal asset allocation strategy."